
#include <glib.h>
#include <assert.h>

#include "priv_city.h"
#include "priv_map.h"
#include "priv_game.h"
#include "log.h"
#include "support.h"
#include "mem.h"
#include "callback.h"
#include "government.h"
#include "net.h"

static struct impr_type improvement_types[B_LAST];
static int improvement_types_count=B_LAST;
static struct city_style *city_styles=NULL;
static int city_styles_count=0;
static GList *cities=NULL;

city_changed_notify_cbt cb_city_changed_notify;
new_city_notify_cbt cb_new_city_notify;
city_destroyed_notify_cbt cb_city_destroyed_notify;
city_construction_finished_notify_cbt cb_city_construction_finished_notify;

void set_city_changed_notify_cb(city_changed_notify_cbt callback)
{
	cb_city_changed_notify=callback;
}

void set_new_city_notify_cb(new_city_notify_cbt callback)
{
	cb_new_city_notify=callback;
}

void set_city_destroyed_notify_cb(city_destroyed_notify_cbt callback)
{
	cb_city_destroyed_notify=callback;
}

void set_city_construction_finished_notify_cb(city_construction_finished_notify_cbt callback)
{
	cb_city_construction_finished_notify=callback;
}

void free_city_styles()
{
	if(!city_styles || !city_styles_count)
		return;

	free(city_styles);
	city_styles=NULL;
	city_styles_count=0;
}

void alloc_city_styles(int size)
{
	city_styles = fc_calloc(size, sizeof(struct city_style));
	city_styles_count=size;
}


void add_improvement_type(struct packet_ruleset_building *p)
{
	struct impr_type *b;

	if(p->id < 0 || p->id >= B_LAST) {
		freelog(LOG_NORMAL,
				"Received bad building id %d in handle_ruleset_building()",
				p->id);
		return;
	}
	b = &improvement_types[p->id];

	sz_strlcpy(b->name, p->name);
	b->tech_req = p->tech_req;
	b->bldg_req = p->bldg_req;
	free(b->terr_gate);
	b->terr_gate = p->terr_gate;		/* pointer assignment */
	free(b->spec_gate);
	b->spec_gate = p->spec_gate;		/* pointer assignment */
	b->equiv_range = p->equiv_range;
	free(b->equiv_dupl);
	b->equiv_dupl = p->equiv_dupl;	/* pointer assignment */
	free(b->equiv_repl);
	b->equiv_repl = p->equiv_repl;	/* pointer assignment */
	b->obsolete_by = p->obsolete_by;
	b->is_wonder = p->is_wonder;
	b->build_cost = p->build_cost;
	b->upkeep = p->upkeep;
	b->sabotage = p->sabotage;
	free(b->effect);
	b->effect = p->effect;		/* pointer assignment */
	b->variant = p->variant;	/* FIXME: remove when gen-impr obsoletes */
	free(b->helptext);
	b->helptext = p->helptext;		/* pointer assignment */

#ifdef DEBUG
	if(p->id == B_LAST-1) {
		int id;
		for (id = 0; id < B_LAST; id++) {
			int inx;
			b = &improvement_types[id];
			freelog(LOG_DEBUG, "Impr: %s...",
					b->name);
			freelog(LOG_DEBUG, "  tech_req    %2d/%s",
					b->tech_req,
					(b->tech_req == A_LAST) ?
					"Never" :
					find_tech_by_id(b->tech_req).name);
			freelog(LOG_DEBUG, "  bldg_req    %2d/%s",
					b->bldg_req,
					(b->bldg_req == B_LAST) ?
					"None" :
					improvement_types[b->bldg_req].name);
			freelog(LOG_DEBUG, "  terr_gate...");
			for (inx = 0; b->terr_gate[inx] != T_LAST; inx++) {
				freelog(LOG_DEBUG, "    %2d/%s",
						b->terr_gate[inx], get_terrain_name(b->terr_gate[inx]));
			}
			freelog(LOG_DEBUG, "  spec_gate...");
			for (inx = 0; b->spec_gate[inx] != S_NO_SPECIAL; inx++) {
				freelog(LOG_DEBUG, "    %2d/%s",
						b->spec_gate[inx], get_special_name(b->spec_gate[inx]));
			}
			freelog(LOG_DEBUG, "  equiv_range %2d/%s",
					b->equiv_range, effect_range_name(b->equiv_range));
			freelog(LOG_DEBUG, "  equiv_dupl...");
			for (inx = 0; b->equiv_dupl[inx] != B_LAST; inx++) {
				freelog(LOG_DEBUG, "    %2d/%s",
						b->equiv_dupl[inx], improvement_types[b->equiv_dupl[inx]].name);
			}
			freelog(LOG_DEBUG, "  equiv_repl...");
			for (inx = 0; b->equiv_repl[inx] != B_LAST; inx++) {
				freelog(LOG_DEBUG, "    %2d/%s",
						b->equiv_repl[inx], improvement_types[b->equiv_repl[inx]].name);
			}
			freelog(LOG_DEBUG, "  obsolete_by %2d/%s",
					b->obsolete_by, advances[b->obsolete_by].name);
			freelog(LOG_DEBUG, "  is_wonder   %2d", b->is_wonder);
			freelog(LOG_DEBUG, "  build_cost %3d", b->build_cost);
			freelog(LOG_DEBUG, "  upkeep      %2d", b->upkeep);
			freelog(LOG_DEBUG, "  sabotage   %3d", b->sabotage);
			freelog(LOG_DEBUG, "  effect...");
			for (inx = 0; b->effect[inx].type != EFT_LAST; inx++) {
				char buf[1024], *ptr;
				ptr = buf;
				my_snprintf(ptr, sizeof(buf)-(ptr-buf), " %d/%s",
							b->effect[inx].type,
							effect_type_name(b->effect[inx].type));
				ptr = strchr(ptr, '\0');
				my_snprintf(ptr, sizeof(buf)-(ptr-buf), " range=%d/%s",
							b->effect[inx].range,
							effect_range_name(b->effect[inx].range));
				ptr = strchr(ptr, '\0');
				my_snprintf(ptr, sizeof(buf)-(ptr-buf), " amount=%d",
							b->effect[inx].amount);
				ptr = strchr(ptr, '\0');
				freelog(LOG_DEBUG, "   %2d. %s", inx, buf);
				ptr = buf;
				my_snprintf(ptr, sizeof(buf)-(ptr-buf), " cond_bldg=%d/%s",
							b->effect[inx].cond_bldg,
							(b->effect[inx].cond_bldg == B_LAST) ?
							"Uncond." :
							improvement_types[b->effect[inx].cond_bldg].name);
				ptr = strchr(ptr, '\0');
				my_snprintf(ptr, sizeof(buf)-(ptr-buf), " cond_gov=%d/%s",
							b->effect[inx].cond_gov,
							(b->effect[inx].cond_gov == game.government_count) ?
							"Uncond." :
							get_government_name(b->effect[inx].cond_gov));
				ptr = strchr(ptr, '\0');
				my_snprintf(ptr, sizeof(buf)-(ptr-buf), " cond_adv=%d/%s",
							b->effect[inx].cond_adv,
							(b->effect[inx].cond_adv == A_NONE) ?
							"Uncond." :
							(b->effect[inx].cond_adv == A_LAST) ?
							"Never" :
							advances[b->effect[inx].cond_adv].name);
				ptr = strchr(ptr, '\0');
				my_snprintf(ptr, sizeof(buf)-(ptr-buf), " cond_eff=%d/%s",
							b->effect[inx].cond_eff,
							(b->effect[inx].cond_eff == EFT_LAST) ?
							"Uncond." :
							effect_type_name(b->effect[inx].cond_eff));
				ptr = strchr(ptr, '\0');
				freelog(LOG_DEBUG, "       %s", buf);
				ptr = buf;
				my_snprintf(ptr, sizeof(buf)-(ptr-buf), " aff_unit=%d/%s",
							b->effect[inx].aff_unit,
							(b->effect[inx].aff_unit == UCL_LAST) ?
							"All" :
							unit_class_name(b->effect[inx].aff_unit));
				ptr = strchr(ptr, '\0');
				my_snprintf(ptr, sizeof(buf)-(ptr-buf), " aff_terr=%d/%s",
							b->effect[inx].aff_terr,
							(b->effect[inx].aff_terr == T_LAST) ?
							"None" :
							(b->effect[inx].aff_terr == T_UNKNOWN) ?
							"All" :
							get_terrain_name(b->effect[inx].aff_terr));
				ptr = strchr(ptr, '\0');
				my_snprintf(ptr, sizeof(buf)-(ptr-buf), " aff_spec=%04X/%s",
							b->effect[inx].aff_spec,
							(b->effect[inx].aff_spec == 0) ?
							"None" :
							(b->effect[inx].aff_spec == S_ALL) ?
							"All" :
							get_special_name(b->effect[inx].aff_spec));
				ptr = strchr(ptr, '\0');
				freelog(LOG_DEBUG, "       %s", buf);
			}
			freelog(LOG_DEBUG, "  variant     %2d", b->variant);	/* FIXME: remove when gen-impr obsoletes */
			freelog(LOG_DEBUG, "  helptext    %s", b->helptext);
		}
	}
#endif
	freelog(LOG_DEBUG, "added '%s' under index %d",p->name,p->id);
}

int improvement_type_exists(int id)
{
	return (find_improvement_type_by_id(id)!=NULL);
}


void add_city_style(struct packet_ruleset_city *p)
{
	int id;
	struct city_style *cs;

	id = p->style_id;
	if (id < 0 || id >= city_styles_count) {
		freelog(LOG_NORMAL, "Received bad citystyle id %d in handle_ruleset_city()",
				id);
		return;
	}
	cs = &city_styles[id];
  
	cs->techreq = p->techreq;
	cs->replaced_by = p->replaced_by;

	sz_strlcpy(cs->name, p->name);
	freelog(LOG_DEBUG, "added '%s' under index %d",p->name,p->style_id);
}

struct city *find_city_by_id(int id)
{
	int i;

	for(i=0;i<g_list_length(cities);i++)
	{
		struct city *p=(struct city *)g_list_nth_data(cities,i);
		if(p->id==id)
			return p;
	}
	return NULL;
}

struct city *find_city_by_position(int x,int y)
{
	int i;

	for(i=0;i<g_list_length(cities);i++)
	{
		struct city *p=(struct city *)g_list_nth_data(cities,i);
		if(p->x==x && p->y==y)
			return p;
	}
	return NULL;
}

int find_cities_by_player(int player_id, int store_size, struct city **store)
{
	int i,found=0;

	for(i=0;i<g_list_length(cities);i++)
	{
		struct city *p=(struct city *)g_list_nth_data(cities,i);
		if(p->owner==player_id)
		{
			store[found]=p;
			found++;
			if(found>=store_size)
				break;
		}
	}
	return found;
}

int can_establish_trade_route(int city1_id, int city2_id)
{
	struct city *pc1=find_city_by_id(city1_id);
	struct city *pc2=find_city_by_id(city2_id);
	int i;
	int r=1;

	if(!pc1 || !pc2 || pc1==pc2 ||
	   (pc1->owner==pc2->owner && 
		map_distance(pc1->x, pc1->y, pc2->x, pc2->y)<=8))
		return 0;
  
	for(i=0;i<4;i++) {
		r*=pc1->trade[i];
		if (pc1->trade[i]==pc2->id) 
			return 0;
	}
	if (r) 
		return 0;
	r = 1;
	for (i=0;i<4;i++) 
		r*=pc2->trade[i];
	return (!r);
}

struct impr_type *find_improvement_type_by_id(int id)
{
	if(id < 0 || id >= improvement_types_count)
		return NULL;
	return &improvement_types[id];
}

int city_got_improvement(int city_id, Impr_Type_id id) 
{
	struct city *pcity=find_city_by_id(city_id);
	
	if (!improvement_type_exists(id))
		return 0;
	else 
		return (pcity->improvements[id]);
}
#define map_get_special(x,y) (find_tile_by_position(x,y)->special)
#define map_get_terrain(x,y) (find_tile_by_position(x,y)->terrain)

static int get_shields_for_tile(int x, int y, struct city *pcity)
{
	int s=0;
	enum tile_special_type spec_t=map_get_special(pcity->x+x-2, pcity->y+y-2);
	enum tile_terrain_type tile_t=map_get_terrain(pcity->x+x-2, pcity->y+y-2);
	struct tile_type *ptt=find_tile_type_by_id(tile_t);
	struct government *g = find_government_by_city(pcity->id);
	int celeb = pcity->is_celebrating;
	int before_penalty = (celeb ? g->celeb_shields_before_penalty
						  : g->shields_before_penalty);

	if (spec_t & S_SPECIAL_1) 
		s=ptt->shield_special_1;
	else if (spec_t & S_SPECIAL_2) 
		s=ptt->shield_special_2;
	else
		s=ptt->shield;

	if (x == 2 &&  y == 2 && !s) /* Always give a single shield on city square */
		s=1;
  
	if (spec_t & S_MINE) {
		s += ptt->mining_shield_incr;
	}
	if (spec_t & S_RAILROAD)
		s+=(s*terrain_control.rail_shield_bonus)/100;
	if (city_affected_by_wonder(pcity, B_RICHARDS))
		s++;
	if (tile_t==T_OCEAN && city_got_improvement(pcity->id, B_OFFSHORE))
		s++;
	/* government shield bonus & penalty */
	if (s)
		s += (celeb ? g->celeb_shield_bonus : g->shield_bonus);
	if (before_penalty && s > before_penalty)
		s--;
	if (spec_t & S_POLLUTION)
		s-=(s*terrain_control.pollution_shield_penalty)/100; /* The shields here is icky */
	return s;
}

static int get_trade_for_tile(int x, int y, struct city *pcity)
{
	enum tile_special_type spec_t=map_get_special(pcity->x+x-2, pcity->y+y-2);
	enum tile_terrain_type tile_t=map_get_terrain(pcity->x+x-2, pcity->y+y-2);
	struct tile_type *ptt=find_tile_type_by_id(tile_t);
	struct government *g = find_government_by_city(pcity->id);
	int t;
 
	if (spec_t & S_SPECIAL_1) 
		t=ptt->trade_special_1;
	else if (spec_t & S_SPECIAL_2) 
		t=ptt->trade_special_2;
	else
		t=ptt->trade;
    
	if ((spec_t & S_RIVER) && (tile_t != T_OCEAN)) {
		t += terrain_control.river_trade_incr;
	}
	if (spec_t & S_ROAD) {
		t += ptt->road_trade_incr;
	}
	if (t) {
		int celeb = pcity->is_celebrating;
		int before_penalty = (celeb ? g->celeb_trade_before_penalty
							  : g->trade_before_penalty);
    
		if (spec_t & S_RAILROAD)
			t+=(t*terrain_control.rail_trade_bonus)/100;

		/* Civ1 specifically documents that Railroad trade increase is before 
		 * Democracy/Republic [government in general now -- SKi] bonus  -AJS */
		if (t)
			t += (celeb ? g->celeb_trade_bonus : g->trade_bonus);

		if(city_affected_by_wonder(pcity, B_COLLOSSUS)) 
			t++;
		if((spec_t&S_ROAD) && city_got_improvement(pcity->id, B_SUPERHIGHWAYS))
			t+=(t*terrain_control.road_superhighway_trade_bonus)/100;
 
		/* government trade penalty -- SKi */
		if (before_penalty && t > before_penalty) 
			t--;
		if (spec_t & S_POLLUTION)
			t-=(t*terrain_control.pollution_trade_penalty)/100; /* The trade here is dirty */
	}
	return t;
}

static int get_food_for_tile(int x, int y, struct city *pcity)
{
	int f;
	enum tile_special_type spec_t=map_get_special(pcity->x+x-2, pcity->y+y-2);
	enum tile_terrain_type tile_t=map_get_terrain(pcity->x+x-2, pcity->y+y-2);
	struct tile_type *ptt=find_tile_type_by_id(tile_t);
	struct government *g = find_government_by_city(pcity->id);
	int celeb = pcity->is_celebrating;
	int before_penalty = (celeb ? g->celeb_food_before_penalty
						  : g->food_before_penalty);
	int city_auto_water;

	city_auto_water = (x==2 && y==2 && 
					   tile_t==ptt->irrigation_result && 
					   terrain_control.may_irrigate);

	if (spec_t & S_SPECIAL_1) 
		f=ptt->food_special_1;
	else if (spec_t & S_SPECIAL_2) 
		f=ptt->food_special_2;
	else
		f=ptt->food;

	if ((spec_t & S_IRRIGATION) || city_auto_water) {
		f += ptt->irrigation_food_incr;
		if (((spec_t & S_FARMLAND) ||
			 (city_auto_water &&
			  player_knows_tech_with_flag(pcity->owner, TF_FARMLAND))) &&
			city_got_improvement(pcity->id, B_SUPERMARKET)) {
			f += (f * terrain_control.farmland_supermarket_food_bonus) / 100;
		}
	}

	if (tile_t==T_OCEAN && city_got_improvement(pcity->id, B_HARBOUR))
		f++;

	if (spec_t & S_RAILROAD)
		f+=(f*terrain_control.rail_food_bonus)/100;

	if (game.civstyle==2 && tile_t==T_MOUNTAINS && x==2 && y==2)
		f++;

	if (f)
		f += (celeb ? g->celeb_food_bonus : g->food_bonus);
	if (before_penalty && f > before_penalty) 
		f--;

	if (spec_t & S_POLLUTION)
		f-=(f*terrain_control.pollution_food_penalty)/100; /* The food here is yucky */

	return f;
}

static int improvement_upkeep_asmiths(struct city *pcity, int i, int asmiths) 
{
	struct impr_type *pimpr=find_improvement_type_by_id(i);
	
	if(!pimpr || pimpr->is_wonder)
		return 0;
	if (asmiths && pimpr->upkeep == 1) 
		return 0;
	return (pimpr->upkeep);
}

static int get_gold_surplus(struct city *pcity)
{
	int asmiths = city_affected_by_wonder(pcity, B_ASMITHS);
	int cost=0;
	int i;

	for (i=0;i<B_LAST;i++) 
		if (city_got_improvement(pcity->id, i)) 
			cost+=improvement_upkeep_asmiths(pcity, i, asmiths);
	return pcity->tax_total-cost;
}

static int get_buy_cost(struct city *pcity)
{
	int total,cost;
	int build=pcity->shield_stock;

	if (pcity->is_building_unit) 
	{
		total=find_unit_type_by_id(pcity->currently_building)->build_cost;
		if (build>=total)
			return 0;
		cost=(total-build)*2+(total-build)*(total-build)/20; 
	} 
	else 
	{
		total=find_improvement_type_by_id(pcity->currently_building)->build_cost;
		if (build>=total)
			return 0;
		cost=(total-build)*2;
		if(pcity->is_building_wonder)
			cost*=2;
	}
	if(!build)
		cost*=2;
	return cost;
}

static int get_turns_to_complete(struct city *pcity, int id, int id_is_unit)
{
	int city_shield_surplus = pcity->shield_surplus;

	if (city_shield_surplus > 0)
	{
		int rounds, cost;
		int shield_stock = pcity->shield_stock;

		if (id_is_unit)
		{
			if (!pcity->is_building_unit)
				shield_stock /= 2;
			cost = find_unit_type_by_id(id)->build_cost;
		}
		else
		{
			if (pcity->is_building_unit ||
				pcity->is_building_wonder != 
				find_improvement_type_by_id(id)->is_wonder)
				shield_stock /= 2;
			cost = find_improvement_type_by_id(id)->build_cost;
		}

		rounds = (cost - shield_stock + city_shield_surplus - 1) /
			city_shield_surplus;

		return (rounds > 0) ? rounds : 1;
	}

	return 999;
}

void update_city(struct packet_city_info *p)
{
	int i, x, y, city_is_new;
	struct city *pcity;
	struct government *g;
	int has_finished_building=0;

	pcity=find_city_by_id(p->id);

	if(pcity)
		assert(pcity->owner == p->owner);

	if(!pcity) {
		city_is_new=1;
		pcity=fc_malloc(sizeof(struct city));
		pcity->id=p->id;
		cities=g_list_append(cities,pcity);
	}
	else 
	{
		city_is_new=0;
		assert(pcity->id == p->id);
	}
  
	pcity->owner=p->owner;
	pcity->x=p->x;
	pcity->y=p->y;
	sz_strlcpy(pcity->name, p->name);
  
	pcity->size=p->size;
	pcity->ppl_happy=p->ppl_happy;
	pcity->ppl_content=p->ppl_content;
	pcity->ppl_unhappy=p->ppl_unhappy;
	pcity->ppl_elvis=p->ppl_elvis;
	pcity->ppl_scientist=p->ppl_scientist;
	pcity->ppl_taxman=p->ppl_taxman;

	pcity->city_options=p->city_options;
  
	for (i=0;i<4;i++)  {
		pcity->trade[i]=p->trade[i];
		pcity->trade_value[i]=p->trade_value[i];
	}
  
	pcity->food_prod=p->food_prod;
	pcity->food_surplus=p->food_surplus;
	pcity->shield_prod=p->shield_prod;
	pcity->shield_surplus=p->shield_surplus;
	pcity->trade_prod=p->trade_prod;
	pcity->corruption=p->corruption;
  
	pcity->luxury_total=p->luxury_total;
	pcity->tax_total=p->tax_total;
	pcity->science_total=p->science_total;
  
	pcity->food_stock=p->food_stock;
	if(pcity->shield_stock > p->shield_stock)
	{
		has_finished_building=1;
	}
	pcity->shield_stock=p->shield_stock;
	pcity->pollution=p->pollution;
    
	pcity->is_building_unit=p->is_building_unit;
	pcity->currently_building=p->currently_building;
#if 0
	if (city_is_new)
		pcity->worklist = create_worklist();
	copy_worklist(pcity->worklist, &p->worklist);
#endif
	pcity->did_buy=p->did_buy;
	pcity->did_sell=p->did_sell;
	pcity->was_happy=p->was_happy;
	pcity->airlift=p->airlift;
  
	i=0;
	for(y=0; y<CITY_MAP_SIZE; y++) {
		for(x=0; x<CITY_MAP_SIZE; x++) {
			pcity->city_map[x][y].type=p->city_map[i++]-'0';
		}
	}
    
	for(i=0; i<B_LAST; i++)
		pcity->improvements[i]=(p->improvements[i]=='1') ? 1 : 0;

	if(city_is_new) {
		freelog(LOG_DEBUG, "New %s city %s id %d (%d %d)",
				find_nation_by_id(find_player_by_id(pcity->owner)
								  ->nation)->name,
				pcity->name, pcity->id, pcity->x, pcity->y);
	}

	/* update calculated values */
	g=find_government_by_city(pcity->id);

	pcity->is_building_wonder=find_improvement_type_by_id(pcity->currently_building)->is_wonder;

	pcity->gold_surplus=get_gold_surplus(pcity);
	pcity->buy_cost=get_buy_cost(pcity);
	pcity->is_happy=(pcity->ppl_happy>=(pcity->size+1)/2 && 
					 pcity->ppl_unhappy==0);
	pcity->is_unhappy=(pcity->ppl_happy<pcity->ppl_unhappy);
	pcity->is_celebrating=(pcity->size >= g->rapture_size && 
						   pcity->was_happy && 
						   pcity->is_happy);
	pcity->is_rapture_growing=(/*pcity->rapture>0 RFFIXME  && */
							   pcity->food_surplus>0 &&
							   government_has_flag(g->id, 
												   G_RAPTURE_CITY_GROWTH));
	pcity->got_citywalls=(city_got_improvement(pcity->id, B_CITY) ||
						  city_affected_by_wonder(pcity, B_WALL));

	for(y=0; y<CITY_MAP_SIZE; y++) {
		for(x=0; x<CITY_MAP_SIZE; x++) {
			pcity->city_map[x][y].food=get_food_for_tile(x,y,pcity);
			pcity->city_map[x][y].trade=get_trade_for_tile(x,y,pcity);
			pcity->city_map[x][y].shields=get_shields_for_tile(x,y,pcity);
		}
	}
	pcity->turns_to_complete=get_turns_to_complete(pcity, 
												   pcity->currently_building,
												   pcity->is_building_unit);


	if(city_is_new)
	{
		CALL(cb_new_city_notify, pcity->id);
	}
	else
	{
		CALL(cb_city_changed_notify, pcity->id);
	}
	if(has_finished_building)
	{
		CALL(cb_city_construction_finished_notify,pcity->id);
	}
}

int city_affected_by_wonder(struct city *pcity, Impr_Type_id id) /*FIX*/
{
	struct city *tmp;
	struct impr_type *pimpr=find_improvement_type_by_id(id);

	assert(improvement_type_exists(id));
	assert(pimpr->is_wonder);
	
	if(wonder_obsolete(id))
		return 0;
	if (city_got_improvement(pcity->id, id))
		return 1;
  
	/* For Manhatten it can be owned by anyone, and it doesn't matter
	 * whether it is destroyed or not.
	 *
	 * (The same goes for Apollo, with respect to building spaceship parts,
	 * but not for getting the map effect.  This function only returns true
	 * for Apollo for the owner of a non-destroyed Apollo; for building
	 * spaceship parts just check (game.global_wonders[id] != 0).
	 * (Actually, this function is not currently used for either Manhatten
	 * or Apollo.))
	 *
	 * Otherwise the player who owns the city needs to have it to
	 * get the effect.
	 */
	if (id==B_MANHATTEN) 
		return (game.global_wonders[id] != 0);
  
	tmp = find_city_by_id(game.global_wonders[id]);
	if (!tmp || tmp->owner!=pcity->owner)
		return 0;
	switch (id) 
	{
		case B_ASMITHS:
		case B_APOLLO:
		case B_CURE:
		case B_GREAT:
		case B_WALL:
		case B_HANGING:
		case B_ORACLE:
		case B_UNITED:
		case B_WOMENS:
		case B_DARWIN:
		case B_LIGHTHOUSE:
		case B_MAGELLAN:
		case B_MICHELANGELO:
		case B_SETI:
		case B_PYRAMIDS:
		case B_LIBERTY:
		case B_SUNTZU:
			return 1;

		case B_ISAAC:
		case B_COPERNICUS:
		case B_SHAKESPEARE:
		case B_COLLOSSUS:
		case B_RICHARDS:
			return 0;

		case B_HOOVER:
		case B_BACH:
			if (pimpr->variant==1) {
				return (find_tile_by_position(tmp->x, tmp->y)->continent ==
						find_tile_by_position(pcity->x, pcity->y)->continent);
			} else {
				return 1;
			}
		default:
			return 0;
	}
} 

int wonder_obsolete(Impr_Type_id id)
{ 
	struct impr_type *pimpr=find_improvement_type_by_id(id);

	if (pimpr->obsolete_by==A_NONE)
		return 0;
	return (game.global_advances[pimpr->obsolete_by]);
}

int can_city_build_unit_type_direct(int city_id, Unit_Type_id id)
{
	struct city *pcity=find_city_by_id(city_id);
	struct unit_type *ptype=find_unit_type_by_id(id);

	if (!ptype)
		return 0;
	if (unit_type_has_flag(id, F_NUCLEAR) && !game.global_wonders[B_MANHATTEN])
		return 0;
	if (get_invention(pcity->owner,ptype->tech_requirement)!=TECH_KNOWN)
		return 0;
	if (!is_terrain_near_tile(pcity->x, pcity->y, T_OCEAN) && 
		ptype->is_water_unit_type)
		return 0;
	return 1;
}

int can_city_build_unit_type(int city_id, Unit_Type_id id)
{
	struct city *pcity=find_city_by_id(city_id);
	
	if (!can_city_build_unit_type_direct(pcity->id, id))
		return 0;
	for(;;)
	{
		id = find_unit_type_by_id(id)->obsoleted_by;
		if(!unit_type_exists(id))
			break;
		if (can_player_build_unit_type_direct(pcity->owner,id))
			return 0;
	}
	return 1;
}

void rename_city(int city_id, char *new_name)
{
	struct packet_city_request packet;

    packet.city_id=city_id;
    packet.worklist.name[0] = '\0';
    sz_strlcpy(packet.name, new_name);
    send_packet_city_request(&aconnection, &packet, PACKET_CITY_RENAME);
}

void buy_city_construction(int city_id)
{
	struct packet_city_request packet;

	assert(find_myself()->economic.gold>=find_city_by_id(city_id)->buy_cost);

	packet.city_id=city_id;
	packet.name[0]='\0';
	packet.worklist.name[0] = '\0';
	send_packet_city_request(&aconnection, &packet, PACKET_CITY_BUY);
}

void set_construction_to_unit(int city_id, Unit_Type_id id)
{
    struct packet_city_request packet;

    packet.city_id=city_id;
    packet.name[0]='\0';
    packet.worklist.name[0] = '\0';
    packet.build_id=id;
    packet.is_build_id_unit_id=1;

    send_packet_city_request(&aconnection, &packet, PACKET_CITY_CHANGE);
}

void set_construction_to_improvement(int city_id,Impr_Type_id id)
{
    struct packet_city_request packet;

    packet.city_id=city_id;
    packet.name[0]='\0';
    packet.worklist.name[0] = '\0';
    packet.build_id=id;
    packet.is_build_id_unit_id=0;

    send_packet_city_request(&aconnection, &packet, PACKET_CITY_CHANGE);
}

void sell_city_improvement(int city_id, int improvement_id)
{
	struct packet_city_request packet;

	packet.city_id=city_id;
	packet.build_id=improvement_id;
	packet.name[0]='\0';
	packet.worklist.name[0] = '\0';
	send_packet_city_request(&aconnection, &packet, PACKET_CITY_SELL);
}
