
#include <glib.h>
#include <assert.h>

#include "priv_unit.h"
#include "mem.h"
#include "log.h"
#include "support.h"
#include "game.h"
#include "callback.h"
#include "net.h"
#include "priv_map.h"

static struct unit_type *unit_types=NULL;
static int unit_types_count=0;
static GList *units=NULL;

caravan_arrived_cbt cb_caravan_arrived;
unit_changed_notify_cbt cb_unit_changed_notify;
new_unit_notify_cbt cb_new_unit_notify;
unit_destroyed_notify_cbt cb_unit_destroyed_notify;

void set_unit_changed_notify_cb(unit_changed_notify_cbt callback)
{
	cb_unit_changed_notify=callback;
}

void set_new_unit_notify_cb(new_unit_notify_cbt callback)
{
	cb_new_unit_notify=callback;
}

void set_unit_destroyed_notify_cb(unit_destroyed_notify_cbt callback)
{
	cb_unit_destroyed_notify=callback;
}

void free_unit_types()
{
	if(!unit_types || !unit_types_count)
		return;

	free(unit_types);
	unit_types=NULL;
	unit_types_count=0;
}

void alloc_unit_types(int size)
{
	unit_types = fc_calloc(size, sizeof(struct unit_type));
	unit_types_count=size;
}

void add_unit_type(struct packet_ruleset_unit *p)
{
	struct unit_type *u;

	if(p->id < 0 || p->id >= unit_types_count || p->id >= U_LAST) {
		freelog(LOG_NORMAL, 
				"Received bad unit_type id %d in handle_ruleset_unit()",
				p->id);
		return;
	}
	u = &unit_types[p->id];

	u->id=p->id;
	sz_strlcpy(u->name, p->name);
	u->move_type          = p->move_type;
	u->build_cost         = p->build_cost;
	u->attack_strength    = p->attack_strength;
	u->defense_strength   = p->defense_strength;
	u->move_rate          = p->move_rate;
	u->tech_requirement   = p->tech_requirement;
	u->vision_range       = p->vision_range;
	u->transport_capacity = p->transport_capacity;
	u->hp                 = p->hp;
	u->firepower          = p->firepower;
	u->obsoleted_by       = p->obsoleted_by;
	u->fuel               = p->fuel;
	u->flags              = p->flags;
	u->roles              = p->roles;
	u->happy_cost         = p->happy_cost;
	u->shield_cost        = p->shield_cost;
	u->food_cost          = p->food_cost;
	u->gold_cost          = p->gold_cost;
	u->paratroopers_range = p->paratroopers_range;
	u->paratroopers_mr_req = p->paratroopers_mr_req;
	u->paratroopers_mr_sub = p->paratroopers_mr_sub;

	free(u->helptext);
	u->helptext = p->helptext;	/* pointer assignment */

	u->is_military=(!unit_type_has_flag(u->id, F_NONMIL));
	u->is_diplomat=(unit_type_has_flag(u->id, F_DIPLOMAT));
	u->is_hiding=(unit_type_has_flag(u->id, F_PARTIAL_INVIS)); /*RFIXME
																 more code*/
	u->is_water_unit_type=(u->move_type == SEA_MOVING);
	u->is_air_unit_type=(u->move_type == AIR_MOVING);
	u->is_heli_unit_type=(u->move_type == HELI_MOVING);
	u->is_ground_unit_type=(u->move_type == LAND_MOVING);

	u->is_ground_units_transport=(u->transport_capacity &&
								  !unit_type_has_flag(u->id, F_MISSILE_CARRIER)
								  && !unit_type_has_flag(u->id, F_CARRIER));
	u->is_air_units_transport=(u->transport_capacity &&
							   (unit_type_has_flag(u->id, F_MISSILE_CARRIER) ||
								unit_type_has_flag(u->id, F_CARRIER)));	

	freelog(LOG_DEBUG,"added '%s' under index %d",p->name,p->id);
}

static int can_help_build_wonder(struct unit *punit, struct city *pcity)
{
	return (unit_has_flag(punit->id, F_CARAVAN) &&
			is_tiles_adjacent(punit->x, punit->y, pcity->x, pcity->y) &&
			punit->owner==pcity->owner && 
			pcity->is_building_wonder  && 
			(find_improvement_type_by_id(pcity->currently_building)->build_cost >
			 pcity->shield_stock));
}

void process_caravan_arrival(struct unit *punit)
{
    if (punit && (punit->can_help_build_wonder_here || 
				  punit->can_est_traderoute_here))
	{
		struct city *pcity_dest = find_city_by_position(punit->x, punit->y);
		struct city *pcity_homecity = find_city_by_id(punit->homecity);
		if (pcity_dest && pcity_homecity) 
		{
			enum caravan_action ca;
			enum packet_type pt;
			struct packet_unit_request req;
			req.unit_id=punit->id;
			req.city_id=pcity_dest->id;
			req.name[0]='\0';
			
			CALL(cb_caravan_arrived, punit->id,&ca);
			if(ca==CA_HELP_BUILDING_WONDER ||
			   ca==CA_EST_TRADE_ROUTE)
			{
				if(ca==CA_HELP_BUILDING_WONDER)
					pt=PACKET_UNIT_HELP_BUILD_WONDER;
				else
					pt=PACKET_UNIT_ESTABLISH_TRADE;

				send_packet_unit_request(&aconnection, &req, pt);
			}
			else if(ca!=CA_KEEP_MOVING)
			{
				assert(0);
			}
		}
    }
}

static int is_a_city_near(int x,int y)
{
	if(game.civstyle==1) 
	{
		return !find_city_by_position(x,y);
	} 
	else 
	{
		int dx,dy;
		
		for(dx=-1; dx<=1; dx++)
			for(dy=-1; dy<=1; dy++)
				if(find_city_by_position(x+dx,y+dy))
					return 0;
	}
	return 1;
}

void calculate_unit_values(struct unit *punit)
{
	int unit_id=punit->id;
	struct city *pcity = find_city_by_position(punit->x, punit->y);
	struct city *phomecity=find_city_by_id(punit->homecity);
	struct unit_type *ptype=find_unit_type_by_id(punit->type);
	struct tile *ptile=find_tile_by_position(punit->x,punit->y);

	punit->is_military=ptype->is_military;
	punit->is_diplomat=ptype->is_diplomat;
	punit->is_hiding=ptype->is_hiding;

	punit->is_water_unit=ptype->is_water_unit_type;
	punit->is_air_unit=ptype->is_air_unit_type;
	punit->is_heli_unit=ptype->is_heli_unit_type;
	punit->is_ground_unit=ptype->is_ground_unit_type;

	punit->can_help_build_wonder_here=(pcity && 
									   can_help_build_wonder(punit, pcity));

	punit->can_est_traderoute_here=(unit_has_flag(unit_id, F_CARAVAN) &&
									phomecity && 
									pcity &&
									can_establish_trade_route(phomecity->id, 
															  pcity->id));
	punit->can_defend_here=(!punit->is_ground_unit && 
							ptile->terrain==T_OCEAN);

	punit->can_paradrop=(unit_has_flag(unit_id, F_PARATROOPERS) &&
						 !punit->paradropped &&
						 punit->moves_left >= ptype->paratroopers_mr_req &&
						 (ptile->special&S_AIRBASE || pcity ));

	punit->can_change_homecity=(pcity && pcity->owner==punit->owner);

	punit->can_unit_do_auto=(unit_has_flag(unit_id, F_SETTLERS) ||
							 (punit->is_military && 
							  pcity));

	punit->can_build_city=(unit_has_flag(unit_id, F_CITIES) &&
						   ptile->terrain!=T_OCEAN &&
						   punit->moves_left &&
						   is_a_city_near(punit->x,punit->y));

	punit->can_add_to_city=(unit_has_flag(unit_id, F_CITIES) &&
							punit->moves_left &&
							pcity &&
							pcity->size <= game.add_to_size_limit &&
							pcity->owner == punit->owner &&
							(!improvement_type_exists(B_AQUEDUCT) ||
							 city_got_improvement(pcity->id, B_AQUEDUCT) ||
							 pcity->size < game.aqueduct_size) &&
							(!improvement_type_exists(B_SEWER) ||
							 city_got_improvement(pcity->id, B_SEWER) ||
							 pcity->size < game.sewer_size));

	punit->is_ground_units_transport=ptype->is_ground_units_transport;
	punit->is_air_units_transport=ptype->is_air_units_transport;
}

void update_unit(struct packet_unit_info *p)
{
	struct city *pcity;
	struct unit *punit;
	int repaint_unit;
	int repaint_city;		/* regards unit's homecity */

	repaint_unit = 0;
	repaint_city = 0;
	punit = find_unit_by_id(p->id);

	if(punit) {
		int dest_x,dest_y;

		if((punit->activity!=p->activity)         /* change activity */
		   || (punit->activity_target!=p->activity_target)) { /*   or
																   act's target */

			punit->activity=p->activity;
			punit->activity_target=p->activity_target;
		}
    
		if(punit->homecity!=p->homecity) { /* change homecity */
			struct city *pcity;
			if((pcity=find_city_by_id(punit->homecity))) {
				/*unit_list_unlink(&pcity->units_supported, punit);*/
			}
			punit->homecity=p->homecity;
			if((pcity=find_city_by_id(punit->homecity))) {
				/*unit_list_insert(&pcity->units_supported, punit);*/
			}
		}

		if(punit->hp!=p->hp) {                      /* hp changed */
			punit->hp=p->hp;
		}
		if (punit->type!=p->type) {
			punit->type=p->type;
		}

		if(punit->x!=p->x || punit->y!=p->y) { /* change position */
			struct city *pcity;

			punit->x=p->x;
			punit->y=p->y;
			punit->fuel=p->fuel;
			punit->hp=p->hp;
      
			pcity=find_city_by_position(punit->x, punit->y);
			if(pcity)  {	
				if(unit_has_flag(punit->id, F_CARAVAN)
				   && punit->owner==game.player_id
				   && (punit->activity!=ACTIVITY_GOTO ||
					   same_pos(punit->goto_dest_x, punit->goto_dest_y,
								pcity->x, pcity->y))
				   && (punit->can_help_build_wonder_here ||
					   punit->can_est_traderoute_here)) {
					process_caravan_arrival(punit);
				}
			}
		}
		if (punit->unhappiness!=p->unhappiness) {
			punit->unhappiness=p->unhappiness;
		}
		if (punit->upkeep!=p->upkeep) {
			punit->upkeep=p->upkeep;
		}
		if (punit->upkeep_food!=p->upkeep_food) {
			punit->upkeep_food=p->upkeep_food;
		}
		if (punit->upkeep_gold!=p->upkeep_gold) {
			punit->upkeep_gold=p->upkeep_gold;
		}

		punit->veteran=p->veteran;
		punit->moves_left=p->movesleft;
		punit->fuel=p->fuel;
		punit->goto_dest_x=p->goto_dest_x;
		punit->goto_dest_y=p->goto_dest_y;
		punit->paradropped=p->paradropped;
		punit->connecting=p->connecting;
  
		dest_x = p->x;
		dest_y = p->y;
		calculate_unit_values(punit);
		CALL(cb_unit_changed_notify, punit->id);
	}
	else 
	{      /* create new unit */
		punit=fc_malloc(sizeof(struct unit));
    
		punit->id=p->id;
		units=g_list_append(units,punit);
    
		punit->owner=p->owner;
		punit->x=p->x;
		punit->y=p->y;
		punit->veteran=p->veteran;
		punit->homecity=p->homecity;
		punit->type=p->type;
		punit->moves_left=p->movesleft;
		punit->hp=p->hp;
		punit->unhappiness=p->unhappiness;
		punit->activity=p->activity;
		punit->upkeep=p->upkeep;
		punit->upkeep_food=p->upkeep_food;
		punit->upkeep_gold=p->upkeep_gold;
		punit->fuel=p->fuel;
		punit->goto_dest_x=p->goto_dest_x;
		punit->goto_dest_y=p->goto_dest_y;
		punit->activity_target=p->activity_target;
		punit->paradropped=p->paradropped;
		punit->connecting=p->connecting;
    
		punit->activity_count=0;	/* never used in client/ or
									   common/ --dwp */
		/*punit->focus_status=FOCUS_AVAIL;*/
    
		freelog(LOG_DEBUG, "New %s %s id %d (%d %d) hc %d %s", 
				find_nation_by_id(find_player_by_id(punit->owner)->nation)->name,
				find_unit_type_by_id(punit->type)->name, 
				punit->x, punit->y, punit->id,
				punit->homecity, (pcity ? pcity->name : "(unknown)"));
		calculate_unit_values(punit);
		CALL(cb_new_unit_notify, punit->id);
	}
}

struct unit *find_unit_by_id(int id)
{
	int i;

	for(i=0;i<g_list_length(units);i++)
	{
		struct unit *p=(struct unit *)g_list_nth_data(units,i);
		if(p->id==id)
			return p;
	}
	return NULL;
}

int unit_exists(int unit_id)
{
	return find_unit_by_id(unit_id)!=NULL;
}

int unit_type_exists(int unit_type_id)
{
	return find_unit_type_by_id(unit_type_id)!=NULL;
}

void show_units()
{
	int i;

	for(i=0;i<g_list_length(units);i++)
	{
		struct unit *p=(struct unit *)g_list_nth_data(units,i);
		freelog(LOG_VERBOSE, "unit: id=%d pos=(%d,%d) type=%d(%s)",
				p->id,p->x,p->y,p->type,find_unit_type_by_id(p->type)->name);
	}	
}

void remove_unit(int id)
{
	if(!unit_exists(id))
		return;

	CALL(cb_unit_destroyed_notify, id);

	units=g_list_remove(units,find_unit_by_id(id));

	assert(!unit_exists(id));
}

struct unit_type *find_unit_type_by_id(Unit_Type_id id)
{
	if(id < 0 || id >= unit_types_count || id >= U_LAST)
		return NULL;
	return &unit_types[id];
}

struct unit_type *find_unit_type_by_name(char *name)
{
	int i;

	for(i=0;i<unit_types_count;i++)
		if(strcmp(unit_types[i].name,name)==0)
			return &unit_types[i];
	return NULL;
}

int build_city(int id)
{
	struct unit *punit=find_unit_by_id(id);
    struct packet_generic_integer req;

	if(!punit->can_build_city)
		return 0;

    req.value = id;
    send_packet_generic_integer(&aconnection,
								PACKET_CITY_NAME_SUGGEST_REQ, &req);
	return 1;
}

int unit_type_has_flag(int unit_type_id,int flag)
{
	assert(unit_type_exists(unit_type_id));
	assert(flag>=0 && flag<F_LAST);

	return BOOL_VAL(find_unit_type_by_id(unit_type_id)->flags & (1<<flag));
}

int unit_has_flag(int unit_id,int flag)
{
	assert(unit_exists(unit_id));

	return unit_type_has_flag(find_unit_by_id(unit_id)->type,flag);
}

int unit_type_has_role(int unit_type_id,int role)
{
	assert(unit_type_exists(unit_type_id));
	assert(role>=L_FIRST && role<L_LAST);

	return BOOL_VAL(find_unit_type_by_id(unit_type_id)->roles & 
					(1<<(role-L_FIRST)));
}

int unit_has_role(int unit_id,int role)
{
	assert(unit_exists(unit_id));

	return unit_type_has_role(find_unit_by_id(unit_id)->type,role);
}

int find_units_by_position(int x, int y, int store_size, 
						   struct unit **units_store)
{
	int i,found=0;

	for(i=0;i<g_list_length(units);i++)
	{
		struct unit *p=(struct unit *)g_list_nth_data(units,i);

		if(p->x==x && p->y==y)
		{
			units_store[found]=p;
			found++;
			if(found>=store_size)
				break;
		}
	}
	return found;	
}

int get_unit_types_count()
{
	return unit_types_count;
}

void move_unit(int unit_id, int x, int y)
{
	struct packet_move_unit move;

	move.unid=unit_id;
	move.x=x;
	move.y=y;

	send_packet_move_unit(&aconnection, &move);
}

int can_unit_do_activity(int unit_id, enum unit_activity activity)
{
	return can_unit_do_activity_targeted(unit_id, activity, 0);
}

static int can_reclaim_ocean(int x, int y)
{
	int i, j, landtiles;

	if (terrain_control.ocean_reclaim_requirement >= 9)
		return 0;

	landtiles = 0;
	for (i = -1; i <= 1; i++) {
		for (j = -1; j <= 1; j++) {
			if (i || j) {
				if (find_tile_by_position(x+i, y+j)->terrain != T_OCEAN) {
					landtiles++;
				}
				if (landtiles >= terrain_control.ocean_reclaim_requirement)
					return 1;
			}
		}
	}

	return 0;
}

int can_unit_do_activity_targeted(int unit_id,
								  enum unit_activity activity, int target)
{
	struct unit *punit;
	struct player *pplayer;
	struct tile *ptile;
	struct tile_type *type;
	int player;


	punit = find_unit_by_id(unit_id);
	player=punit->owner;
	pplayer = find_player_by_id(player);
	ptile = find_tile_by_position(punit->x, punit->y);
	type = find_tile_type_by_id(ptile->terrain);

	switch(activity) 
	{
		case ACTIVITY_IDLE:
		case ACTIVITY_GOTO:
			return 1;

		case ACTIVITY_POLLUTION:
			return unit_has_flag(unit_id, F_SETTLERS) && 
				(ptile->special&S_POLLUTION);

		case ACTIVITY_ROAD:
			return terrain_control.may_road &&
				unit_has_flag(unit_id, F_SETTLERS) &&
				!(ptile->special&S_ROAD) && type->road_time &&
				((ptile->terrain!=T_RIVER && !(ptile->special&S_RIVER)) || 
				 player_knows_tech_with_flag(player, TF_BRIDGE));

		case ACTIVITY_MINE:
			/* Don't allow it if someone else is irrigating this tile.
			 * *Do* allow it if they're transforming - the mine may survive */
			if (terrain_control.may_mine &&
				unit_has_flag(unit_id, F_SETTLERS) &&
				( (ptile->terrain==type->mining_result && 
				   !(ptile->special&S_MINE)) ||
				  (ptile->terrain!=type->mining_result &&
				   type->mining_result!=T_LAST &&
				   (ptile->terrain!=T_OCEAN ||
					can_reclaim_ocean(punit->x, punit->y)) &&
				   (type->mining_result!=T_OCEAN ||
					!(find_city_by_position(punit->x, punit->y)))) )) 
			{
				struct unit *punits[10];
				int units,i;
				
				units=find_units_by_position(punit->x, punit->y,10,punits);
				for(i=0;i<units;i++)
				{
					if(punits[i]->activity==ACTIVITY_IRRIGATE) 
						return 0;
				}
				return 1;
			}
			else 
				return 0;

		case ACTIVITY_IRRIGATE:
			/* Don't allow it if someone else is mining this tile.
			 * *Do* allow it if they're transforming - the irrigation
			 * *may survive */
			if (terrain_control.may_irrigate &&
				unit_has_flag(unit_id, F_SETTLERS) &&
				(!(ptile->special&S_IRRIGATION) ||
				 (!(ptile->special&S_FARMLAND) &&
				  player_knows_tech_with_flag(player, TF_FARMLAND))) &&
				( (ptile->terrain==type->irrigation_result && 
				   is_irrigation_water_adjacent(punit->x, punit->y)) ||
				  (ptile->terrain!=type->irrigation_result &&
				   type->irrigation_result!=T_LAST &&
				   (ptile->terrain!=T_OCEAN ||
					can_reclaim_ocean(punit->x, punit->y)) &&
				   (type->irrigation_result!=T_OCEAN ||
					!(find_city_by_position(punit->x, punit->y)))) )) 
			{
				struct unit *punits[10];
				int units,i;
				
				units=find_units_by_position(punit->x, punit->y,10,punits);
				for(i=0;i<units;i++)
				{
					if(punits[i]->activity==ACTIVITY_MINE)
						return 0;
				}
				return 1;
			} 
			else 
				return 0;

		case ACTIVITY_FORTIFYING:
			return punit->is_ground_unit &&
				(punit->activity != ACTIVITY_FORTIFIED) &&
				!unit_has_flag(unit_id, F_SETTLERS);

		case ACTIVITY_FORTIFIED:
			return 0;

		case ACTIVITY_FORTRESS:
			return unit_has_flag(unit_id, F_SETTLERS) &&
				player_knows_tech_with_flag(player, TF_FORTRESS) &&
				!(ptile->special&S_FORTRESS) && ptile->terrain!=T_OCEAN;

		case ACTIVITY_AIRBASE:
			return unit_has_flag(unit_id, F_AIRBASE) &&
				player_knows_tech_with_flag(player, TF_AIRBASE) &&
				!(ptile->special&S_AIRBASE) && ptile->terrain!=T_OCEAN;

		case ACTIVITY_SENTRY:
			return 1;

		case ACTIVITY_RAILROAD:
			/* if the tile has road, the terrain must be ok.. */
			return terrain_control.may_road &&
				unit_has_flag(unit_id, F_SETTLERS) &&
				((ptile->special&S_ROAD) || 
				 (punit->connecting
				  && (type->road_time &&
					  ((ptile->terrain!=T_RIVER && !(ptile->special&S_RIVER))
					   || player_knows_tech_with_flag(player, TF_BRIDGE)))))
				&& !(ptile->special&S_RAILROAD) &&
				player_knows_tech_with_flag(player, TF_RAILROAD);

		case ACTIVITY_PILLAGE:
			{
				int pspresent;
				int psworking;

				pspresent = get_tile_infrastructure_set(ptile->special);
				if (pspresent && punit->is_ground_unit) {
					psworking = get_unit_tile_pillage_set(punit->x, punit->y);
					if (target == S_NO_SPECIAL)
						return ((pspresent & (~psworking)) != 0);
					else if ((game.civstyle != 2) &&
							 (target != get_preferred_pillage(pspresent)))
						return 0;
					else
						return ((pspresent & (~psworking) & target) != 0);
				} else {
					return 0;
				}
			}

		case ACTIVITY_EXPLORE:
			return (punit->is_ground_unit || punit->is_water_unit);

		case ACTIVITY_TRANSFORM:
			return terrain_control.may_transform &&
				(type->transform_result!=T_LAST) &&
				(ptile->terrain!=type->transform_result) &&
				(ptile->terrain!=T_OCEAN ||
				 can_reclaim_ocean(punit->x, punit->y)) &&
				(type->transform_result!=T_OCEAN ||
				 !(find_city_by_position(punit->x, punit->y))) &&
				unit_has_flag(unit_id, F_TRANSFORM);

		default:
			freelog(LOG_NORMAL,"Unknown activity %d\n",activity);
			return 0;
	}
}

void set_unit_activity(int unit_id, enum unit_activity new_activity)
{
	struct packet_unit_info info;
	struct unit *punit=find_unit_by_id(unit_id);

	info.id=unit_id;
	info.x=punit->x;
	info.y=punit->y;
	info.homecity=punit->homecity;
	info.veteran=punit->veteran;
	info.type=punit->type;
	info.movesleft=punit->moves_left;
	info.activity=new_activity;
	info.activity_target=0;
	info.select_it=0;
	
	send_packet_unit_info(&aconnection, &info);
}

/* from server/unittools.c: can_unit_move_to_tile */
int can_move_unit(int unit_id, int dest_x, int dest_y)
{
	struct unit *punit=find_unit_by_id(unit_id);
	int src_x = punit->x;
	int src_y = punit->y;
	struct tile *pfromtile,*ptotile;
#if 0
	int zoc;
	struct city *pcity;
#endif

	if(map_move_cost(unit_id, dest_x, dest_y) > punit->moves_left &&
	   punit->moves_left < punit->move_rate)
		return 0;

	/* 1) */
	if (punit->activity!=ACTIVITY_IDLE
		&& punit->activity!=ACTIVITY_GOTO
		&& !punit->connecting)
		return 0;

	/* 2) */
	if (dest_x<0 || dest_x>=map.xsize || dest_y<0 || dest_y>=map.ysize)
		return 0;

	/* 3) */
	if (!is_tiles_adjacent(src_x, src_y, dest_x, dest_y))
		return 0;

	pfromtile = find_tile_by_position(src_x, src_y);
	ptotile = find_tile_by_position(dest_x, dest_y);

	if (punit->is_ground_unit) {
		/* 5) */
		if (ptotile->terrain==T_OCEAN 
#if 0 /* RFFIXME */
			&& ground_unit_transporter_capacity(dest_x, 
												dest_y, punit->owner) <= 0)
			return 0;

		/* Moving from ocean */
		if (pfromtile->terrain==T_OCEAN) {
			/* 6) */
			if (!unit_has_flag(unit_id, F_MARINES)
				&& is_enemy_city_tile(ptotile, punit->owner)) 
			{
				return 0;
			}
		}
	} else if (punit->is_water_unit) {
		/* 7) */
		if (ptotile->terrain!=T_OCEAN
			&& ptotile->terrain!=T_UNKNOWN
			&& !is_allied_city_tile(ptotile, punit->owner))
			return 0;
	}

	/* 4) */
	if (is_non_allied_unit_tile(ptotile, punit->owner))
		return 0;

	/* 8) */
	if (is_non_attack_unit_tile(ptotile, punit->owner)) {
		return 0;
	}

	/* 9) */
	pcity = find_city_by_position(dest_x,dest_y);
	if (pcity && players_non_attack(pcity->owner, punit->owner)) {
		return 0;
	}

	/* 10) */
	zoc = ignore_zoc || zoc_ok_move(punit, dest_x, dest_y);

	return zoc;
#else
			)
			return 0;
    }
return 1;
#endif
}
