/********************************************************************** 
 Freeciv - Copyright (C) 1996 - A Kjeldberg, L Gregersen, P Unold
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/
#ifndef FC__PLAYER_H
#define FC__PLAYER_H

#include "nation.h"
#include "tech.h"
#include "unit.h"
#include "city.h"

#define PLAYER_DEFAULT_TAX_RATE 0
#define PLAYER_DEFAULT_SCIENCE_RATE 100
#define PLAYER_DEFAULT_LUXURY_RATE 0

enum handicap_type {
  H_NONE=0, /* no handicaps */
  H_RIGIDPROD=1, /* can't switch to/from building_unit without penalty */
  H_MAP=2, /* only knows map_get_known tiles */
  H_TECH=4, /* doesn't know what enemies have researched */
  H_CITYBUILDINGS=8, /* doesn't know what buildings are in enemy cities */
  H_CITYUNITS=16, /* doesn't know what units are in enemy cities */
  H_STACKS=32, /* doesn't know what units are in stacks */
  H_VETERAN=64, /* doesn't know veteran status of enemy units */
  H_SUB=128, /* doesn't know where subs may be lurking */
/* below this point are milder handicaps that I can actually implement -- Syela */
  H_RATES=256, /* can't set its rates beyond government limits */
  H_TARGETS=512, /* can't target anything it doesn't know exists */
  H_HUTS=1024 /* doesn't know which unseen tiles have huts on them */
/* anything else I have forgotten?  Let me know. -- Syela */
};


struct player_economic {
  int gold;
  int tax;
  int science;
  int luxury;
};

struct player_research {
	int researched;     /* # bulbs reseached */
	int researchpoints; /* # bulbs to complete */
	int researching;    /* invention being researched in */
	int tech_last_researched;
	int researched_techs;
	int tech_goal;
	int future_tech;
	unsigned char inventions[A_LAST];
};

struct player_score {
  int happy;
  int content;
  int unhappy;
  int taxmen;
  int scientists;
  int elvis;
  int wonders;
  int techs;
  int techout;
  int landarea;
  int settledarea;
  int population;
  int cities;
  int units;
  int pollution;
  int literacy;
  int bnp;
  int mfg;
  int spaceship;
};

/* Diplomatic states (how one player views another).
 * (Some diplomatic states are "pacts" (mutual agreements), others aren't.)
 */
enum diplstate_type {
  DS_NEUTRAL = 0,
  DS_WAR,
  DS_CEASEFIRE,
  DS_PEACE,
  DS_ALLIANCE,
  DS_NO_CONTACT,
  DS_LAST	/* leave this last */
};

struct player_diplstate {
  enum diplstate_type type;	/* this player's disposition towards other */
  /* the following are for "pacts" */
  int turns_left;		/* until pact (e.g., cease-fire) ends */
  int has_reason_to_cancel;	/* 0: no, 1: this turn, 2: this or next turn */
};

/**************************************************************************
  command access levels for client-side use; at present, they are
  only used for server commands typed at the client chatline
 **************************************************************************/
enum cmdlevel_id {    /* access levels for users to issue commands        */
  ALLOW_NONE = 0,     /* user may issue no commands at all                */
  ALLOW_INFO,         /* user may issue informational commands            */
  ALLOW_CTRL,         /* user may issue commands that affect game & users */
  ALLOW_HACK,         /* user may issue *all* commands - dangerous!       */

  ALLOW_NUM,          /* the number of levels                             */
  ALLOW_UNRECOGNIZED  /* used as a failure return code                    */
};
/*  the set command is a special case:                                    */
/*    - ALLOW_CTRL is required for SSET_TO_CLIENT options                 */
/*    - ALLOW_HACK is required for SSET_TO_SERVER options                 */

/***************************************************************************
  On the distinction between nations(formerly races), players, and users,
  see freeciv_hackers_guide.txt
***************************************************************************/

struct player {
  int id;
  char name[MAX_LEN_NAME];
  char username[MAX_LEN_NAME];
  int is_male;
  int government;
  Nation_Type_id nation;
  int turn_done;
  int nturns_idle;
  int is_alive;
  int got_tech;
  int revolution;
  int capital; /* bool used to give player capital in first city. */
  int embassy;
  int reputation;
  struct player_diplstate diplstates[MAX_NUM_PLAYERS + MAX_NUM_BARBARIANS];
  int city_style;
	/*struct unit_list units;
	  struct city_list cities;*/
  struct player_score score;
  struct player_economic economic;
  struct player_research research;
	/*struct player_spaceship spaceship;*/
  int is_connected;
  struct connection *conn;
  char addr[MAX_LEN_ADDR];
};

struct player *find_myself();
struct player *find_player_by_id(int player_id);
struct player *find_player_by_name(char *name);
struct player *find_player_by_name_prefix(const char *name,
					  enum m_pre_result *result);
struct player *find_player_by_user(char *name);
int player_exists(int player_id);

int player_knows_tech_with_flag(int player_id, int flag);
int get_invention(int player_id, int tech_id);
/* Callbacks */

typedef void (*choose_science_cbt)(int *tech_to_research_store,
								   int *tech_goal_store);
void set_choose_science_cb(choose_science_cbt callback);

typedef void (*choose_government_cbt)(int *government_id_store);
void set_choose_government_cb(choose_government_cbt callback);

/* Unused functions */

#if 0
int player_has_embassy(struct player *pplayer, struct player *pplayer2);

int player_can_see_unit(struct player *pplayer, struct unit *punit);
int player_owns_city(struct player *pplayer, struct city *pcity);

int player_in_city_radius(struct player *pplayer, int x, int y);
int player_owns_active_wonder(int player_id,Impr_Type_id id);
int player_owns_active_government_change_wonder(int player_id);
int player_knows_improvement_tech(struct player *pplayer,
								  Impr_Type_id id);

void player_limit_to_government_rates(struct player *pplayer);

struct city *find_palace(struct player *pplayer);

const char *reputation_text(const int rep);
const char *diplstate_text(const enum diplstate_type type);

/* we have an int in some contexts, a pointer in others.  Yuk! */
const struct player_diplstate *pplayer_get_diplstate(const struct player *pplayer,
						     const struct player *pplayer2);
const struct player_diplstate *player_get_diplstate(const int player,
						    const int player2);
int pplayers_at_war(const struct player *pplayer, 
		    const struct player *pplayer2);
int players_at_war(const int player, const int player2);
int pplayers_allied(const struct player *pplayer,
		    const struct player *pplayer2);
int players_allied(const int player, const int player2);
int pplayers_non_attack(const struct player *pplayer,
			const struct player *pplayer2);
int players_non_attack(const int player, const int player2);

const char *cmdlevel_name(enum cmdlevel_id lvl);
enum cmdlevel_id cmdlevel_named(const char *token);

int is_barbarian(const struct player *pplayer);

int tech_goal_turns(struct player *plr, int goal);
int get_next_tech(struct player *plr, int goal);
#endif

int player_can_change_to_government(int player_id, int government);

#endif  /* FC__PLAYER_H */
