/********************************************************************** 
 Freeciv - Copyright (C) 1996 - A Kjeldberg, L Gregersen, P Unold
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/
#ifndef FC__UNIT_H
#define FC__UNIT_H

#include "shared.h"

struct city;
struct government;

typedef int Unit_Type_id;
  
#define U_LAST MAX_NUM_ITEMS
/*
  U_LAST is a value which is guaranteed to be larger than all
  actual Unit_Type_id values.  It is used as a flag value;
  it can also be used for fixed allocations to ensure able
  to hold full number of unit types.
*/

enum unit_activity {
	ACTIVITY_IDLE, ACTIVITY_POLLUTION, ACTIVITY_ROAD, ACTIVITY_MINE,
	ACTIVITY_IRRIGATE, ACTIVITY_FORTIFIED, ACTIVITY_FORTRESS, ACTIVITY_SENTRY,
	ACTIVITY_RAILROAD, ACTIVITY_PILLAGE, ACTIVITY_GOTO, ACTIVITY_EXPLORE,
	ACTIVITY_TRANSFORM, ACTIVITY_UNKNOWN, ACTIVITY_AIRBASE, ACTIVITY_FORTIFYING,
	ACTIVITY_LAST   /* leave this one last */
};

enum unit_move_type {
	LAND_MOVING = 1, SEA_MOVING, HELI_MOVING, AIR_MOVING
};

/* Classes for unit types.
 * (These must correspond to unit_class_names[] in unit.c.)
 */
enum unit_class_id {
	UCL_AIR,
	UCL_HELICOPTER,
	UCL_LAND,
	UCL_MISSILE,
	UCL_NUCLEAR,
	UCL_SEA,
	UCL_LAST	/* keep this last */
};

typedef enum unit_class_id Unit_Class_id;

enum unit_focus_status {
	FOCUS_AVAIL, FOCUS_WAIT, FOCUS_DONE  
};

enum diplomat_actions {
	DIPLOMAT_BRIBE, DIPLOMAT_EMBASSY, DIPLOMAT_SABOTAGE,
	DIPLOMAT_STEAL, DIPLOMAT_INCITE, SPY_POISON, 
	DIPLOMAT_INVESTIGATE, SPY_SABOTAGE_UNIT,
	SPY_GET_SABOTAGE_LIST,
	DIPLOMAT_ANY_ACTION   /* leave this one last */
};

enum diplomat_client_actions {
	DIPLOMAT_CLIENT_POPUP_DIALOG
};

/* Unit "special effects" flags:
   Note this is now an enumerated type, and not power-of-two integers
   for bits, though unit_type.flags is still a bitfield, and code
   which uses unit_flag() without twiddling bits is unchanged.
   (It is easier to go from i to (1<<i) than the reverse.)
*/
enum unit_flag_id { 
	F_CARAVAN=0,
	F_MISSILE,   
	F_IGZOC,     
	F_NONMIL,      
	F_IGTER,       
	F_CARRIER,     
	F_ONEATTACK,   
	F_PIKEMEN,     
	F_HORSE,       
	F_IGWALL,      
	F_FIELDUNIT,   
	F_AEGIS,       
	F_FIGHTER,     
	F_MARINES,     
	F_PARTIAL_INVIS,    /* Invisibile except when adjacent (Submarine) */   
	F_SETTLERS,         /* Does not include ability to found cities */
	F_DIPLOMAT,    
	F_TRIREME,          /* Trireme sinking effect */
	F_NUCLEAR,          /* Nuclear attack effect */
	F_SPY,              /* Enhanced spy abilities */
	F_TRANSFORM,        /* Can transform terrain types (Engineers) */
	F_PARATROOPERS,
	F_AIRBASE,          /* Can build Airbases */
	F_CITIES,           /* Can build cities */
	F_IGTIRED,          /* Ignore tired negative bonus when attacking */
	F_MISSILE_CARRIER,  /* Like F_CARRIER, but missiles only (Submarine) */
	F_NO_LAND_ATTACK,   /* Cannot attack vs land squares (Submarine) */
	F_LAST
};

/* Unit "roles": these are similar to unit flags but differ in that
   they don't represent intrinsic properties or abilities of units,
   but determine which units are used (mainly by the server or AI)
   in various circumstances, or "roles".
   Note that in some cases flags can act as roles, eg, we don't need
   a role for "settlers", because we can just use F_SETTLERS.
   (Now have to consider F_CITIES too)
   So we make sure flag values and role values are distinct,
   so some functions can use them interchangably.
*/
#define L_FIRST 64		/* should be >= F_LAST */
enum unit_role_id {
	L_FIRSTBUILD=L_FIRST, /* is built first when city established */
	L_EXPLORER,           /* initial explorer unit */
	L_HUT,                /* can be found in hut */
	L_HUT_TECH,           /* can be found in hut, global tech required */
	L_PARTISAN,           /* is created in Partisan circumstances */
	L_DEFEND_OK,          /* ok on defense (AI) */
	L_DEFEND_GOOD,        /* primary purpose is defense (AI) */
	L_ATTACK_FAST,        /* quick attacking unit (Horse..Armor) (unused)*/
	L_ATTACK_STRONG,      /* powerful attacking unit (Catapult..) (unused) */
	L_FERRYBOAT,	        /* is useful for ferrying (AI) */
	L_BARBARIAN,          /* barbarians unit, land only */
	L_BARBARIAN_TECH,     /* barbarians unit, global tech required */
	L_BARBARIAN_BOAT,     /* barbarian boat */
	L_BARBARIAN_BUILD,    /* what barbarians should build */
	L_BARBARIAN_BUILD_TECH, /* barbarians build when global tech */
	L_BARBARIAN_LEADER,   /* barbarian leader */
	L_BARBARIAN_SEA,      /* sea raider unit */
	L_BARBARIAN_SEA_TECH, /* sea raider unit, global tech required */
	L_LAST
};

struct unit_type {
	int id;
	char name[MAX_LEN_NAME];
	char name_orig[MAX_LEN_NAME];	      /* untranslated */
	enum unit_move_type move_type;
	int build_cost;
	int attack_strength;
	int defense_strength;
	int move_rate;
	int tech_requirement;
	int vision_range;
	int transport_capacity;
	int hp;
	int firepower;
	int obsoleted_by;
	int fuel;

	unsigned int flags;
	unsigned int roles;

	int happy_cost;  /* unhappy people in home city */
	int shield_cost; /* normal upkeep cost */
	int food_cost;   /* settler food cost */
	int gold_cost;   /* gold upkeep (n/a now, maybe later) */

	int paratroopers_range; /* only valid for F_PARATROOPERS */
	int paratroopers_mr_req;
	int paratroopers_mr_sub;

	char *helptext;

	/* calculated values */

	int is_military;
	int is_diplomat;
	int is_hiding;

	int is_water_unit_type;
	int is_air_unit_type;
	int is_heli_unit_type;
	int is_ground_unit_type;

	int is_ground_units_transport;
	int is_air_units_transport;
};

struct unit {
	int id;
	int owner;
	int x, y;
	int veteran;
	int homecity;
	Unit_Type_id type;
	int moves_left;
	int hp;
	int activity;
	int activity_count;
	int unhappiness;
	int upkeep;
	int upkeep_food;
	int upkeep_gold;
	int fuel;
	int goto_dest_x, goto_dest_y;
	int activity_target;
	int paradropped;
	int connecting;
	int carried;
	int select_it;

	/* calculated values */

	int can_help_build_wonder_here;
	int can_est_traderoute_here;
	int can_defend_here;
	int can_paradrop;
	int can_change_homecity;
	int can_unit_do_auto;

	int is_military;
	int is_diplomat;
	int is_hiding;

	int is_water_unit;
	int is_air_unit;
	int is_heli_unit;
	int is_ground_unit;

	int is_ground_units_transport;
	int is_air_units_transport;

	int can_build_city;
	int can_add_to_city;

	int move_rate;
};

/* Unit type functions */

struct unit_type *find_unit_type_by_id(Unit_Type_id id);
struct unit_type *find_unit_type_by_name(char *name);
int unit_type_exists(int unit_type_id);
int get_unit_types_count();

int unit_type_has_flag(Unit_Type_id id, int flag);
int unit_type_has_role(Unit_Type_id id, int role);

/* Unit functions */

struct unit *find_unit_by_id(int id);
int find_units_by_position(int x, int y, int store_size, 
						   struct unit **units_store);
int unit_exists(int unit_id);

int unit_has_flag(int unit_id, int flag);
int unit_has_role(int unit_id, int role);

int can_unit_do_activity(int unit_id, enum unit_activity activity);
int can_unit_do_activity_targeted(int unit_id,
								  enum unit_activity activity, int target);

void set_unit_activity(int unit_id, enum unit_activity new_activity);

/* 0 for failure */
int build_city(int unit_id);

void move_unit(int unit_id, int x, int y);
int can_move_unit(int unit_id, int dest_x, int dest_y);

/* Callbacks */

enum caravan_action {CA_HELP_BUILDING_WONDER=1, CA_EST_TRADE_ROUTE,
					 CA_KEEP_MOVING};
typedef void (*caravan_arrived_cbt)(int caravan_id, 
									enum caravan_action *action_store);
void set_caravan_arrived_cb(caravan_arrived_cbt callback);

typedef void (*unit_changed_notify_cbt)(int unit_id);
void set_unit_changed_notify_cb(unit_changed_notify_cbt callback);

typedef void (*new_unit_notify_cbt)(int unit_id);
void set_new_unit_notify_cb(new_unit_notify_cbt callback);

typedef void (*unit_destroyed_notify_cbt)(int unit_id);
void set_unit_destroyed_notify_cb(unit_destroyed_notify_cbt callback);

/* Unused functions */

#if 0
int unit_can_help_build_wonder(struct unit *punit, struct city *pcity);
int unit_can_airlift_to(struct unit *punit, struct city *pcity);

int can_unit_do_connect(struct unit *punit, enum unit_activity activity);
int can_unit_do_activity(struct unit *punit, enum unit_activity activity);
int can_unit_do_activity_targeted(struct unit *punit,
								  enum unit_activity activity, int target);
void set_unit_activity(struct unit *punit, enum unit_activity new_activity);
void set_unit_activity_targeted(struct unit *punit,
								enum unit_activity new_activity, 
								int new_target);

int is_unit_activity_on_tile(enum unit_activity activity, int x, int y);

int is_ground_threat(int player_id, struct unit *punit);
int is_square_threatened(int player_id, int x, int y);

int diplomat_can_do_action(struct unit *pdiplomat,
						   enum diplomat_actions action, 
						   int destx, int desty);
int is_diplomat_action_available(struct unit *pdiplomat,
                                 enum diplomat_actions action,
                                 int destx, int desty);

int ground_unit_transporter_capacity(int x, int y, int playerid);
int missile_carrier_capacity(int x, int y, int playerid);
int airunit_carrier_capacity(int x, int y, int playerid);

int utype_shield_cost(struct unit_type *ut, struct government *g);
int utype_food_cost(struct unit_type *ut, struct government *g);
int utype_happy_cost(struct unit_type *ut, struct government *g);
int utype_gold_cost(struct unit_type *ut, struct government *g);

int can_upgrade_unittype(int player_id, Unit_Type_id id);
int unit_upgrade_price(int player_id, Unit_Type_id from, Unit_Type_id to);

void role_unit_precalcs(void);
int num_role_units(int role);
Unit_Type_id get_role_unit(int role, int index);
Unit_Type_id best_role_unit(struct city *pcity, int role);
#endif

#endif  /* FC__UNIT_H */
